import math
from hashlib import sha3_256

SECURITY_PARAMETER = 256
# Note that notably `MAXIMUM_HASH_DIFFICULTY` and `SPACE_TO_PROVE_IN_BITS` should be sent from the prover to the verifier before the second verifier initialization step. However to ease this prototype this communication isn't performed.
# For the ease of this prototype only 1 / (2 ** n) values are supported for `MAXIMUM_HASH_DIFFICULTY`.
MAXIMUM_HASH_DIFFICULTY = 1 / 16
# The following space is a minimum.
INITIAL_SPACE_TO_PROVE_IN_BITS = 10_000
SPACE_TO_PROVE_IN_BITS = 10_000

LEAST_DIFFICULT_HASH = int('F' * (SECURITY_PARAMETER // 4), 16)

NUMBER_OF_BITS_IN_AVERAGE_FOR_SOLVING_PUZZLE = int(math.log2(1 / MAXIMUM_HASH_DIFFICULTY))

PROBABILITY_TO_CATCH_MALICIOUS = 1 - 2 ** -50
# Let's hardcode that we dedicate 1% of extra storage for the proof probability.
SPACE_UPPER_BOUND_IN_BITS_POTENTIALLY_MALICIOUS = math.ceil(SPACE_TO_PROVE_IN_BITS / 100)

PROBABILITY_NOT_TO_CATCH_MALICIOUS_WITH_ONE_QUERY = (SPACE_TO_PROVE_IN_BITS - SPACE_UPPER_BOUND_IN_BITS_POTENTIALLY_MALICIOUS) / SPACE_TO_PROVE_IN_BITS

NUMBER_OF_BITS_TO_VERIFY = math.ceil(math.log(1 - PROBABILITY_TO_CATCH_MALICIOUS) / math.log(PROBABILITY_NOT_TO_CATCH_MALICIOUS_WITH_ONE_QUERY))
NUMBER_OF_ENTRIES_TO_VERIFY = math.ceil(NUMBER_OF_BITS_TO_VERIFY / NUMBER_OF_BITS_IN_AVERAGE_FOR_SOLVING_PUZZLE)

SPACE_TO_PROVE_IN_BITS = math.ceil(SPACE_TO_PROVE_IN_BITS + SPACE_UPPER_BOUND_IN_BITS_POTENTIALLY_MALICIOUS)

ENTRIES_NUMBER = math.ceil(SPACE_TO_PROVE_IN_BITS / NUMBER_OF_BITS_IN_AVERAGE_FOR_SOLVING_PUZZLE)
# To ease Merkle tree implementation use a power of 2 number of leafs.
ENTRIES_NUMBER = 2 ** math.ceil(math.log2(ENTRIES_NUMBER))
LEVELS = int(math.log2(ENTRIES_NUMBER)) + 1

def hash(x):
    return sha3_256(x.encode('ascii')).hexdigest()

# Storage could be done by using a separator but is it optimal? In fact how to do a separator optimally in binary? Could for instance encode on two bits, for normal bits the first one should be 0 otherwise if the first is 1 then whatever the second bit, it's a separator.
# TODO: sure that this function is meaningly correct? cf below comment, should also consider the serialization of the actual array - should in a similar way verify `ENTRIES_NUMBER` computation
# Have to pay attention that `0` isn't interesting us.
def numberOfBitsOf(n):
    return math.floor(math.log2(n)) + 1

def otherIndex(n):
    return n + (1 if n % 2 == 0 else -1)

def doesHashMatchDifficulty(hashed):
    hashedInteger = int(hashed, 16)
    hashedDifficulty = hashedInteger / LEAST_DIFFICULT_HASH
    return hashedDifficulty <= MAXIMUM_HASH_DIFFICULTY
