#!/usr/bin/python3

## Proof of Space-Time verifier

import secrets, common, ast, sys

## Initialization phase: Generate a random bitstring to send to the prover.
# To make sure that the following protocol execution can't be used partially or totally by the prover to reduce the cost of another protocol execution.

protocolInitializationPhaseId = secrets.token_hex(common.SECURITY_PARAMETER // 8) if len(sys.argv) == 1 else sys.argv[1]
print(f'{protocolInitializationPhaseId=}')

while True:
    print()
    ## Execution phase: 1. Generate a random bitstring to send to the prover.
    # TODO: improve step numbering, as it's a bit unclear what step of each to run after which one
    # Wait the initialization phase termination from the prover side before starting the execution phase.

    protocolExecutionPhaseId = secrets.token_hex(common.SECURITY_PARAMETER // 8) if len(sys.argv) <= 2 else sys.argv[2]
    print(f'{protocolExecutionPhaseId=}')

    ## Execution phase: 2. Receive the prover commitment to the table contents given this random challenge.

    merkleTreeRoot = input('merkleTreeRoot: ')

    ## Execution phase: 3. Generate and send a random set of indexes to the prover.

    # Need to know the actual table size to choose a random set of indexes. This point isn't discussed to the best of my knowledge in Simple Proofs of Space-Time and Rational Proofs of Storage, Tal Moran and Ilan Orlov, 2016. In a more important manner, it's unclear how the prover proves to the verifier precisely with high probability or *for sure* that he is actually storing a given number of bits.
    # For this prototype the prover claims a given table size and the verifier estimate the total table number of bits thanks to the requested set of its entries.

    # Note that there's no advantage to request the last table entry by hoping that the prover is storing the nonces with delta in such order, for instance he can have stored the delta the other way around which screw up this additional request aim.

    indexesRequest = [] if len(sys.argv) <= 3 else ast.literal_eval(sys.argv[3])

    if indexesRequest == []:
        while len(indexesRequest) < common.NUMBER_OF_ENTRIES_TO_VERIFY:
            index = secrets.randbelow(common.ENTRIES_NUMBER)
            if not common.otherIndex(index) in indexesRequest and not index in indexesRequest:
                indexesRequest += [index]

    indexesRequestFilePath = 'indexesRequest.txt'

    with open(indexesRequestFilePath, 'w') as f:
        f.write(str(indexesRequest))

    print(f'indexesRequest: {indexesRequestFilePath}')

    ## Execution phase: 4. Receive and verify corresponding table entries and commitment openings.

    # Add mentions of rationality and code its factor? or/and precise when to run initialization and execution phases
    # TODO: precising `entries` size could be interesting.

    # Proceeding in a minimum number of passes could be interesting but this can be left for future work.

    with open(input('entries: ')) as f:
        entries = ast.literal_eval(f.read())

    nonces = set()
    for entry in entries:
        if type(entry) == str:
            break
        if entry in nonces:
            print('A given nonce was used several times!')
            exit(1)
        nonces.add(entry)

    merkleTreeLevels = [{} for _ in range(common.LEVELS)]

    for index in indexesRequest:
        otherIndex = common.otherIndex(index)
        leftIndex, rightIndex = sorted([index, otherIndex])
        for i, nonce in zip([leftIndex, rightIndex], [entries.pop(0) for _ in range(2)]):
            hashed = common.hash(protocolInitializationPhaseId + str(nonce))
            if not common.doesHashMatchDifficulty(hashed):
                print("The received difficulty doesn't match the claimed one!")
                exit(1)
            hashed = common.hash(protocolInitializationPhaseId + protocolExecutionPhaseId + str(nonce))
            merkleTreeLevels[0][i] = hashed

    for merkleTreeLevelsIndex, merkleTreeLevel in enumerate(merkleTreeLevels[:-1]):
        for index in merkleTreeLevel.copy():
            otherIndex = common.otherIndex(index)
            leftIndex, rightIndex = sorted([index, otherIndex])
            if merkleTreeLevelsIndex > 0:
                merkleTreeLevels[merkleTreeLevelsIndex][otherIndex] = entries.pop(0)
            leftHash, rightHash = [merkleTreeLevels[merkleTreeLevelsIndex][i] for i in [leftIndex, rightIndex]]
            merkleTreeLevels[merkleTreeLevelsIndex + 1][leftIndex // 2] = common.hash(leftHash + rightHash)

    # Note that `actualStorage` is the average storage that the prover should have dedicated to be able to solve this execution phase challenge.
    # If any entry corresponding to requested indexes is incorrect, the Merkle tree won't match or the verification of the claimed difficulty of the provided hash will fail.
    actualStorage = common.ENTRIES_NUMBER * common.NUMBER_OF_BITS_IN_AVERAGE_FOR_SOLVING_PUZZLE

    if merkleTreeRoot == merkleTreeLevels[-1][0] and actualStorage >= common.SPACE_TO_PROVE_IN_BITS:
        print(f'Verified with probability {common.PROBABILITY_TO_CATCH_MALICIOUS} that the prover is storing in average {common.INITIAL_SPACE_TO_PROVE_IN_BITS} (actually the prover is storing in average {actualStorage} bits)')
    else:
        print("The prover hasn't provided a correct proof!")

